"""DamageScanner - a directe damage assessment toolkit

Sensitivity analysis using the damagescanner
"""

import os
import pandas
import numpy

from SALib.sample import morris
import SALib.analyze.morris

from tqdm import tqdm
import matplotlib.pyplot as plt

import rasterio

from damagescanner.core import RasterScanner

def run(landuse_map,inun_map,curve_path,maxdam_path,save=False,**kwargs):
    """
    Perform sensitivity analysis of the results. Currently only works with the RasterScanner.
    
     Arguments:
        *landuse_map* : Shapefile, Pandas DataFrame or Geopandas GeoDataFrame 
        with land-use information of the area.
     
        *inun_map* : GeoTiff with inundation depth per grid cell. Make sure 
        that the unit of the inundation map corresponds with the unit of the 
        first column of the curves file. 
     
        *curve_path* : File with the stage-damage curves of the different 
        land-use classes. Can also be a pandas DataFrame (but not a numpy Array).
     
        *maxdam_path* : File with the maximum damages per land-use class 
        (in euro/m2). Can also be a pandas DataFrame (but not a numpy Array).
        
    Returns:
        *Figure* : Functions returns a figure with the spread in losses and 
        a spyder plot with the influence of each parameter.

    """
    
    # Set up problem for sensitivity analysis
    problem = {
          'num_vars': 3,
          'names': ['inundation', 'maxdam','curves'],
          'bounds': [[0.5,1.5],[0.5,1.5],[1,5]]}
    
    # And create parameter values
    param_values = morris.sample(problem, 10, num_levels=5, grid_jump=2,local_optimization =True)
    
    # Run analysis with the specified parameter values
    collect_losses = []
    for param_set in tqdm(param_values,total=len(param_values)):
        with rasterio.open(inun_map) as src:
            inundation = src.read()[0,:,:]*param_set[0]
        inundation[inundation>10] = 0
        inundation[inundation<0] = 0
        
        maxdam = pandas.read_csv(maxdam_path,skiprows=1).values*param_set[1]
        
        curves = pandas.read_csv(curve_path).values
        curves[:,1:] = curves[:,1:]/param_set[2]
        
        loss_df = RasterScanner(landuse_map,inundation,curves,maxdam,save=False,
                                                         in_millions=True)
        collect_losses.append(loss_df[0].sum().values[0])
        
    # Obtain results
    Si = SALib.analyze.morris.analyze(problem, numpy.array(param_values), 
                                          numpy.array(collect_losses),
                                     print_to_console=False, grid_jump=2, num_levels=5)
    
    # And plot the results
    fig = plt.figure(figsize=(10, 5))
    ax1 = plt.subplot(121)
    ax2 = plt.subplot(122, projection='polar')
    
    color_scheme_map = ['#264653','#2A9D8F']
        
    # Plot boxplot with spread of total losses
    ax1.boxplot(numpy.array(collect_losses),showfliers=False, patch_artist=True,
                boxprops=dict(facecolor=color_scheme_map[1], edgecolor='black'),
                capprops=dict(color='black'),
                whiskerprops=dict(color='black'),
                flierprops=dict(color=color_scheme_map[1], markeredgecolor='black'),
                medianprops=dict(color=color_scheme_map[1]),
                )
    
    ax1.set_title('Spread of total losses',fontsize=18,fontweight='black', y=1.08)
    
    # Plot spyder plot with influence of different variables
    risk_sens = pandas.DataFrame.from_dict(Si)
    risk_sens['rel'] = abs(risk_sens['mu'])/abs(risk_sens['mu']).sum()*100
    risk_sens['mu'] = risk_sens['mu']
    risk_sens = risk_sens.groupby('names').sum()
    risk_sens = risk_sens.T
    risk_sens.columns = ['inundation', 'maxdam', 'curve']      
    
    stats=risk_sens.loc['rel',numpy.array(risk_sens.columns)].values
    
    angles=numpy.linspace(0, 2*numpy.pi, len(numpy.array(risk_sens.columns)), endpoint=False)
    
    # close the plot
    stats=numpy.concatenate((stats,[stats[0]]))
    angles=numpy.concatenate((angles,[angles[0]]))
    
    ax2.plot(angles, stats, 'o-', linewidth=2,color=color_scheme_map[0])
    ax2.set_ylim([0, 100])   
    ax2.fill(angles, stats, alpha=0.25,color=color_scheme_map[0])
    ax2.set_thetagrids(angles * 180/numpy.pi, numpy.array(risk_sens.columns))
    ax2.tick_params(axis='x',labelsize=14,labelcolor='black',color='black',) # pad=12
    
    ax2.set_title('Influence of variables',fontsize=18,fontweight='black', y=1.08)
    
    fig.tight_layout()
    
    # Save results if desired
    if save:
        output_path = kwargs['output_path']
        if not os.path.exists(output_path):
            os.mkdir(output_path)
        fig.savefig(os.path.join(output_path,'sensitivity_analysis.png'),dpi=300)
    


# -*- coding: utf-8 -*-
"""
Created on Thu Jan 31 11:54:09 2019

@author: cenv0574
"""
import os
import pandas
from scipy import integrate
from tqdm import tqdm

from damagescanner.core import RasterScanner,VectorScanner

def monetary_risk(RPS,loss_list):
    """
    """
    return integrate.simps(y=loss_list[::-1], x=RPS[::-1])

def RasterBased(landuse_ras,inundation_path,curve_path,maxdam_path,per_landuse=False):
    """
    Raster-based implementation of a risk assessment.
    
    Arguments:
        *landuse_map* : GeoTiff with land-use information per grid cell. Make sure 
        the land-use categories correspond with the curves and maximum damages 
        (see below). Furthermore, the resolution and extend of the land-use map 
        has to be exactly the same as the inundation map.
     
        *inun_map* : GeoTiff with inundation depth per grid cell. Make sure 
        that the unit of the inundation map corresponds with the unit of the 
        first column of the curves file.
     
        *curve_path* : File with the stage-damage curves of the different 
        land-use classes. Can also be a pandas DataFrame or numpy Array.
     
        *maxdam_path* : File with the maximum damages per land-use class 
        (in euro/m2). Can also be a pandas DataFrame or numpy Array.

    Optional Arguments:
        *per_landuse* : Set to **True** if you would like the output er land-use class.

    Returns:
        *total_risk* : Returns the total risk for the area
        
    """    
    
    # Get list of inundation maps and make sure they are in ascending order
    list_inundation_maps = os.listdir(inundation_path)
    list_inundation_maps.sort(key=lambda f: int(''.join(filter(str.isdigit, f))))
    list_inundation_maps = [os.path.join(inundation_path,x) for x in list_inundation_maps]
    
    RPS = [1/int(''.join(filter(str.isdigit, x))) for x in list_inundation_maps]
    return_periods =  ['rp_'+''.join(filter(str.isdigit, x)) for x in list_inundation_maps]

    # loop through the inundation maps and calculate the losses for each map
    damage_list = {}
    for inun_map in tqdm(list_inundation_maps,total=len(list_inundation_maps),desc='Return Periods'):
        return_period = 'rp_'+''.join(filter(str.isdigit, inun_map))
        damage_list[return_period] = RasterScanner(landuse_ras,inun_map,curve_path,maxdam_path)[0]

    damage_table = pandas.concat(damage_list).unstack(level=0)
    damage_table.columns = damage_table.columns.droplevel(0)
    damage_table = damage_table[return_periods]

    # calculate the total risk and return it
    if per_landuse:
        return pandas.DataFrame(damage_table.apply(lambda x: monetary_risk(RPS,list(x)),axis=1),columns=['tot_risk'])
    else:    
        return monetary_risk(RPS,list(damage_table.sum(axis=0)))/1e6


def VectorBased(landuse_vec,inundation_path,curve_path,maxdam_path,per_landuse=False):
    """
    Vector based implementation of a direct damage assessment
    
    Arguments:
        *landuse_map* : Shapefile, Pandas DataFrame or Geopandas GeoDataFrame 
        with land-use information of the area.
     
        *inun_map* : GeoTiff with inundation depth per grid cell. Make sure 
        that the unit of the inundation map corresponds with the unit of the 
        first column of the curves file. 
     
        *curve_path* : File with the stage-damage curves of the different 
        land-use classes. Can also be a pandas DataFrame (but not a numpy Array).
     
        *maxdam_path* : File with the maximum damages per land-use class 
        (in euro/m2). Can also be a pandas DataFrame (but not a numpy Array).

    Optional Arguments:
        *per_landuse* : Set to **True** if you would like the output er land-use class.
        
        *landuse_col* : Specify the column name of the unique landuse id's. 
        Default is set to **landuse**.
        
    Returns:
        *total_risk* : Returns the total risk for the area
    """
     
    # Get list of inundation maps and make sure they are in ascending order
    list_inundation_maps = os.listdir(inundation_path)
    list_inundation_maps.sort(key=lambda f: int(''.join(filter(str.isdigit, f))))
    list_inundation_maps = [os.path.join(inundation_path,x) for x in list_inundation_maps]
    
    RPS = [1/int(''.join(filter(str.isdigit, x))) for x in list_inundation_maps]
    return_periods =  ['rp_'+''.join(filter(str.isdigit, x)) for x in list_inundation_maps]

    # loop through the inundation maps and calculate the losses for each map
    damage_list = {}
    for inun_map in tqdm(list_inundation_maps,total=len(list_inundation_maps),desc='Return Periods'):
        return_period = 'rp_'+''.join(filter(str.isdigit, inun_map))
        damage_list[return_period] =  pandas.DataFrame(
            VectorScanner(landuse_vec,inun_map,curve_path,maxdam_path,print_tqdm=False).groupby('landuse')['tot_dam'].sum())      
        
    damage_table = pandas.concat(damage_list).unstack(level=0)
    damage_table.columns = damage_table.columns.droplevel(0)
    damage_table = damage_table[return_periods]

    # calculate the total risk and return it
    if per_landuse:
        return pandas.DataFrame(damage_table.apply(lambda x: monetary_risk(RPS,list(x)),axis=1),columns=['tot_risk'])
    else:    
        return monetary_risk(RPS,list(damage_table.sum(axis=0)))/1e6


"""DamageScanner - a directe damage assessment toolkit

Plot results of DamageScanner
"""
import os
import numpy
import rasterio
import pandas
import geopandas
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.patches import Patch
import matplotlib.pyplot as plt
from rasterio.plot import show

def landuse_vector(landuse,color_dict={},save=False,**kwargs):
    """
    Arguments:
        *landuse_map* : Shapefile, Pandas DataFrame or Geopandas GeoDataFrame 
        with land-use information of the area.

    Optional Arguments:
        *color_dict* : Supply a dictionary with the land-use classes as keys 
        and color hex codes as values. If empty, it will use a default list (based on OSM attributes)
        
        *save* : Set to True if you would like to save the output. Requires 
        several **kwargs**
        
    kwargs:
        *output_path* : Specify where files should be saved.
        
        *scenario_name*: Give a unique name for the files that are going to be saved.
    """
    # load land-use map
    if isinstance(landuse,str):
        landuse = geopandas.read_file(landuse)
    elif isinstance(landuse, geopandas.GeoDataFrame):
        landuse = landuse.copy()
    elif isinstance(landuse, pandas.DataFrame):
        landuse = geopandas.GeoDataFrame(landuse,geometry='geometry')
    else:
        print('ERROR: landuse should either be a shapefile, a GeoDataFrame or a pandas Dataframe with a geometry column')
    
    if len(color_dict) == 0:
        color_dict = {  "grass":'#c3eead',
                        "forest":'#1c7426',
                        "orchard":'#fe6729',
                        "residential":'#f13013',
                        "industrial":'#0f045c',
                        "retail":'#b71456',
                        "farmland":'#fcfcb9',
                        "cemetery":'#c39797',
                        "construction":'#c0c0c0',
                        "meadow":'#c3eead',
                        "farmyard":'#fcfcb9',
                        "scrub":'#98574d',
                        "allotments":'#fbffe2',
                        "reservoir":'#8af4f2',
                        "static_caravan":'#ff3a55',
                        "commercial":'#27142c'}
                    
    map_dict = dict(zip(color_dict.keys(),[x for x in range(len(color_dict))]))

    landuse['int_landuse'] = landuse.landuse.apply(lambda x: map_dict[x])

    fig, ax = plt.subplots(1, 1,figsize=(12,10))
    color_scheme_map = list(color_dict.values())

    cmap = LinearSegmentedColormap.from_list(name='landuse',
                                         colors=color_scheme_map)  


    landuse.plot(column='int_landuse',ax=ax,linewidth=0,cmap=cmap)

    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_axis_off()
    legend_elements = []
    for iter_,item in enumerate(color_dict):
        legend_elements.append(Patch(facecolor=color_scheme_map[iter_],label=item))        

    ax.legend(handles=legend_elements,edgecolor='black',facecolor='#fefdfd',prop={'size':12},loc=(1.02,0.2)) 

    if save:
        if 'output_path' in kwargs:
            output_path = kwargs['output_path']
            if not os.path.exists(output_path):
                os.mkdir(output_path)
        if 'scenario_name' in kwargs:
            scenario_name = kwargs['scenario_name']
        fig.tight_layout()
        fig.savefig(os.path.join(output_path,'landuse_{}.png'.format(scenario_name)),dpi=350, bbox_inches='tight')
 
    return ax

def landuse_raster(landuse_ras,color_dict={},save=False,**kwargs):
    """
    Arguments:
        *landuse_map* : path to GeoTiff with land-use information per grid cell. 

    Optional Arguments:
        *color_dict* : Supply a dictionary with the land-use classes as keys 
        and color hex codes as values. If empty, it will use a default list (which will probably fail)
        
        *save* : Set to True if you would like to save the output. Requires 
        several **kwargs**
        
    kwargs:
        *output_path* : Specify where files should be saved.
        
        *scenario_name*: Give a unique name for the files that are going to be saved.
    """
    fig, ax = plt.subplots(1, 1,figsize=(12,10))

    if len(color_dict) == 0:
        color_dict = {
        110 : '#fb897e' ,     111 : '#b40e3e' ,     112 : '#ee0000' ,     120 : '#ee0000' , 
        130 : '#edc3c3' ,     131 : '#d97489' ,     133 : '#da6c99' ,     134 : '#da6c99' , 
        135 : '#da6c99' ,     136 : '#da6c99' ,     140 : '#331e36' ,     141 : '#363b74' , 
        142 : '#363b74' ,     143 : '#363b74' ,     144 : '#363b74' ,     150 : '#69868A' , 
        151 : '#363b74' ,     152 : '#363b74' ,     210 : '#fcfcb9' ,     211 : '#fcfcb9' , 
        220 : '#B59E99' ,     221 : '#7b323d' ,     230 : '#c3eead' ,     240 : '#b29600' , 
        250 : '#b29600' ,     310 : '#89ec46' ,     320 : '#89ec46' ,     330 : '#89ec46' , 
        340 : '#89ec46' ,     350 : '#89ec46' ,     360 : '#1c7426' ,     370 : '#e7d7bf' , 
        380 : '#3bbbb3' ,     410 : '#010000' ,     430 : '#71818e' ,     440 : '#c0c0c0' , 
        450 : '#c0c0c0' ,     460 : '#c0c0c0' ,     520 : '#545454' ,     530 : '#c39797' , 
        550 : '#7b323d' ,     560 : '#c0c0c0' ,    630 : '#b0e0e6' ,      640 : '#b0e0e6' , 
        911 : '#b0e0e6' 
        }

    with rasterio.open(landuse_ras) as src:
        landuse = src.read()[0,:,:]
        transform = src.transform

        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_axis_off()
        color_scheme_map = list(color_dict.values())

        cmap = LinearSegmentedColormap.from_list(name='landuse',
                                             colors=color_scheme_map)  

        map_dict = dict(zip(color_dict.keys(),[x for x in range(len(color_dict))]))
        # Function to be vectorized
        def map_func(val, dictionary):
            return dictionary[val] if val in dictionary else val 

        # Vectorize map_func
        vfunc  = numpy.vectorize(map_func)

        # Run
        landuse = vfunc(landuse, map_dict)

        if 'background' in kwargs:
            show(landuse,ax=ax,cmap=cmap,transform=transform,alpha=0.5)
        else:
            show(landuse,ax=ax,cmap=cmap,transform=transform)
 
    if save:
        if 'output_path' in kwargs:
            output_path = kwargs['output_path']
            if not os.path.exists(output_path):
                os.mkdir(output_path)
        if 'scenario_name' in kwargs:
            scenario_name = kwargs['scenario_name']

        fig.tight_layout()
        fig.savefig(os.path.join(output_path,'landuse_{}.png'.format(scenario_name)),dpi=350, bbox_inches='tight')


    return ax


def inundation_map(inun_map,lu_raster=False,lu_vector=False,save=False,**kwargs):
    """
    Arguments:
        *inun_map* : GeoTiff with inundation depth per grid cell.

    Optional Arguments:
        *lu_raster* : Set to **True** if you would like to use the landuse raster 
        as background.
            
        *lu_vector* : Set to **True** if you would like to use the landuse raster 
        as background.
        
        *save* : Set to **True** if you would like to save the output. Requires 
        several **kwargs**
        
    kwargs:
        *landuse_map* : Path to land-use map.
        
        *output_path* : Specify where files should be saved.
        
        *scenario_name*: Give a unique name for the files that are going to be saved.
        
    """    
    if (lu_raster == False) & (lu_vector == False):
        fig, ax = plt.subplots(1, 1,figsize=(12,10))

        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_axis_off()
        
    elif (lu_raster == False) & (lu_vector == True):
        landuse = kwargs['landuse_map']
        ax = landuse_vector(landuse)

    elif (lu_raster == True) & (lu_vector == False):
        landuse = kwargs['landuse_map']
        ax = landuse_raster(landuse)
        
    with rasterio.open(inun_map) as src:
        inundation = numpy.array(src.read()[0,:,:],dtype=float)
        inundation[inundation == 0] = numpy.nan

        max_value = 6
        show(inundation,ax=ax,cmap='Blues',transform=src.transform,zorder=2,vmax=max_value)

    if save:
        if 'output_path' in kwargs:
            output_path = kwargs['output_path']
            if not os.path.exists(output_path):
                os.mkdir(output_path)
        if 'scenario_name' in kwargs:
            scenario_name = kwargs['scenario_name']

        fig.tight_layout()
        fig.savefig(os.path.join(output_path,'inundation_{}.png'.format(scenario_name)),dpi=350, bbox_inches='tight')

     
def damagemap_vector(losses,bins=[],save=False,**kwargs):
    """
    Arguments:
        *losses* : Shapefile, Pandas DataFrame or Geopandas GeoDataFrame 
        with land-use information of the area.

    Optional Arguments:
        *bins* : Supply list of bin values for the colorscheme. If empty, it will use a default list.
        
        *save* : Set to True if you would like to save the output. Requires 
        several **kwargs**.
        
    kwargs:
        *output_path* : Specify where files should be saved.
        
        *scenario_name*: Give a unique name for the files that are going to be saved.
    """

    fig, ax = plt.subplots(1, 1,figsize=(12,10))
    color_scheme_map =  ['white','#fee5d9','#fcae91','#fb6a4a','#de2d26','#a50f15']

    cmap = LinearSegmentedColormap.from_list(name='damages',
                                         colors=color_scheme_map)  
    labels = [0,1,2,3,4]
    if len(bins) == 0:
        bins = [0,10000,100000,500000,1e6,losses.tot_dam.max()]

    losses['binned'] = pandas.cut(losses['tot_dam'], bins=bins, labels=labels)
    losses['binned'] = losses['binned'].fillna(0)

    losses.plot(column='binned',ax=ax,linewidth=0.1,cmap=cmap,edgecolor='black')

    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_axis_off()
    legend_elements = []
    for iter_,item in enumerate(bins):
        if iter_ < len(bins)-1:
            legend_elements.append(Patch(facecolor=color_scheme_map[iter_+1],label='{}-{} Euro'.format(int(bins[iter_]),int(bins[iter_+1]))))        
       

    ax.legend(handles=legend_elements,edgecolor='black',facecolor='#fefdfd',prop={'size':12},loc=(1.02,0.4)) 

    if save:
        if 'output_path' in kwargs:
            output_path = kwargs['output_path']
            if not os.path.exists(output_path):
                os.mkdir(output_path)
        if 'scenario_name' in kwargs:
            scenario_name = kwargs['scenario_name']

        fig.tight_layout()
        fig.savefig(os.path.join(output_path,'Famagemap_{}.png'.format(scenario_name)),dpi=350, bbox_inches='tight')
        
def damagemap_raster(damagemap,landuse,lu_raster=False,bins=[],save=False,**kwargs):
    """
    Arguments:
        *damagemap* : Numpy array of loss output.

        *landuse_map* : path to GeoTiff with land-use information per grid cell. 
        

    Optional Arguments:
        *lu_raster* : Set to **True** if you would like to use the landuse raster 
        as background.        
        
        *bins* : Supply list of bin values for the colorscheme. If empty, it will use a default list.
        
        *save* : Set to True if you would like to save the output. Requires 
        several **kwargs**.
        
    kwargs:
        *output_path* : Specify where files should be saved.
        
        *scenario_name*: Give a unique name for the files that are going to be saved.
    """

    if (lu_raster == False):
        fig, ax = plt.subplots(1, 1,figsize=(12,10))

        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_axis_off()
        
    else:
        ax = landuse_raster(landuse,background=True)

    with rasterio.open(landuse) as src:
        damagemap = numpy.array(damagemap,dtype=float)
        damagemap[damagemap == 0] = numpy.nan

        color_scheme_map =  ['#fee5d9','#fcbba1','#fc9272','#fb6a4a','#de2d26','#a50f15']

        cmap = LinearSegmentedColormap.from_list(name='damages',
                                             colors=color_scheme_map) 
        if len(bins) == 0:
            bins = [1,1000,5000,10000,50000,100000]

        show(damagemap,ax=ax,cmap=cmap,transform=src.transform,zorder=2)

        legend_elements = []
        for iter_,item in enumerate(bins):
            if iter_ < len(bins)-1:
                legend_elements.append(Patch(facecolor=color_scheme_map[iter_],label='{}-{} Euro'.format(int(bins[iter_]),int(bins[iter_+1]))))        
            else:
                legend_elements.append(Patch(facecolor=color_scheme_map[iter_],label='> {} Euro'.format(int(bins[iter_]))))        

        ax.legend(handles=legend_elements,edgecolor='black',facecolor='#fefdfd',prop={'size':12},loc=(1.02,0.4)) 
                  

    if save:
        if 'output_path' in kwargs:
            output_path = kwargs['output_path']
            if not os.path.exists(output_path):
                os.mkdir(output_path)
        if 'scenario_name' in kwargs:
            scenario_name = kwargs['scenario_name']

        fig.tight_layout()
        fig.savefig(os.path.join(output_path,'Damagemap_{}.png'.format(scenario_name)),dpi=350, bbox_inches='tight')
        



"""DamageScanner - a directe damage assessment toolkit

Copyright (C) 2019 Elco Koks. All versions released under the MIT license.
"""
import pkg_resources

__author__ = "Elco Koks"
__copyright__ = "Elco Koks"
__license__ = "MIT"

try:
    __version__ = pkg_resources.get_distribution(__name__).version
except Exception:
    __version__ = 'unknown'
    
__all__ = ['core','plot','vector','raster','sensitivity','utils']



import shapely
import geopandas
import pandas
from osgeo import ogr,gdal
import os
import numpy as np
from tqdm import tqdm
import pyproj


def query_b(geo_type, key_col, **val_constraint):
    """
    Builds an SQL query from the values passed to the function.
    Arguments:
            geo_type : Type of geometry (osm layer) to search for.
            key_col : A list of keys/columns that should be selected from the layer.
            val_constraint : A dictionary of constraints for the values. e.g. WHERE 'value'>20 or 'value'='constraint'
    Returns:
        string: A SQL query string.
    """
    query = "SELECT osm_id"
    
    for column in key_col:
        query += ", " + column

    query += " FROM " + geo_type + " WHERE "

    if val_constraint:
        constraint_clauses = []
        for key, values in val_constraint.items():
            for value in values:
                constraint_clauses.append(f"{key} {value}")
        query += " AND ".join(constraint_clauses) + " AND "
        
    query += f"{key_col[0]} IS NOT NULL"
    return query


def retrieve(osm_path, geo_type, key_col, **val_constraint):
    """
    Function to extract specified geometry and keys/values from OpenStreetMap
    Arguments:
        *osm_path* : file path to the .osm.pbf file of the region 
        for which we want to do the analysis.     
        *geo_type* : Type of Geometry to retrieve. e.g. lines, multipolygons, etc.
        *key_col* : These keys will be returned as columns in the dataframe.
        ***val_constraint: A dictionary specifying the value constraints.  
        A key can have multiple values (as a list) for more than one constraint for key/value.  
    Returns:
        *GeoDataFrame* : a geopandas GeoDataFrame with all columns, geometries, and constraints specified.    
    """
    driver = ogr.GetDriverByName('OSM')
    data = driver.Open(osm_path)
    query = query_b(geo_type, key_col, **val_constraint)
    sql_lyr = data.ExecuteSQL(query)
    features = []

    # cl = columns 
    cl = ['osm_id'] 
    for a in key_col: cl.append(a)
    if data is not None:
        for feature in sql_lyr:
            try:
                if feature.GetField(key_col[0]) is not None:
                    shapely_geo = shapely.from_wkt(feature.geometry().ExportToWkt()) 
                    if shapely_geo is None:
                        continue
                    # field will become a row in the dataframe.
                    field = []
                    for i in cl: field.append(feature.GetField(i))
                    field.append(shapely_geo)   
                    features.append(field)
            except:
                print("WARNING: skipped OSM feature")   
    else:
        print("ERROR: Nonetype error when requesting SQL. Check required.")    
    cl.append('geometry')                   
    if len(features) > 0:
        return geopandas.GeoDataFrame(features, columns=cl, crs={'init': 'epsg:4326'})
    else:
        print("WARNING: No features or No Memory. returning empty GeoDataFrame") 
        return geopandas.GeoDataFrame(columns=['osm_id','geometry'], crs={'init': 'epsg:4326'})

def landuse(osm_path):
    """
    Function to extract land-use polygons from OpenStreetMap    
    Arguments:
        *osm_path* : file path to the .osm.pbf file of the region 
        for which we want to do the analysis.        
    Returns:
        *GeoDataFrame* : a geopandas GeoDataFrame with all unique land-use polygons.    
    """    
    return(retrieve(osm_path,'multipolygons',['landuse']))

def buildings(osm_path):
    """
    Function to extract building polygons from OpenStreetMap    
    Arguments:
        *osm_path* : file path to the .osm.pbf file of the region 
        for which we want to do the analysis.        
    Returns:
        *GeoDataFrame* : a geopandas GeoDataFrame with all unique building polygons.    
    """
    return retrieve(osm_path, 'multipolygons',['building','amenity'])

def roads(osm_path):
    """
    Function to extract road linestrings from OpenStreetMap  
    Arguments:
        *osm_path* : file path to the .osm.pbf file of the region 
        for which we want to do the analysis.        
    Returns:
        *GeoDataFrame* : a geopandas GeoDataFrame with all unique road linestrings.
    """   
    return retrieve(osm_path,'lines',['highway']) 
 
def railway(osm_path):
    """
    Function to extract railway linestrings from OpenStreetMap   
    Arguments:
        *osm_path* : file path to the .osm.pbf file of the region 
        for which we want to do the analysis.       
    Returns:
        *GeoDataFrame* : a geopandas GeoDataFrame with all unique land-use polygons.
    """ 
    return retrieve(osm_path,'lines',['railway','service'],**{"service":[" IS NOT NULL"]})

def ferries(osm_path):
    """
    Function to extract road linestrings from OpenStreetMap
    Arguments:
        *osm_path* : file path to the .osm.pbf file of the region 
        for which we want to do the analysis.
    Returns:
        *GeoDataFrame* : a geopandas GeoDataFrame with all unique road linestrings.
    """
    return retrieve(osm_path,'lines',['route'],**{"route":["='ferry'",]})

def electricity(osm_path):
    """
    Function to extract railway linestrings from OpenStreetMap    
    Arguments:
        *osm_path* : file path to the .osm.pbf file of the region 
        for which we want to do the analysis.        
    Returns:
        *GeoDataFrame* : a geopandas GeoDataFrame with all unique land-use polygons.   
    """    
    return retrieve(osm_path,'lines',['power','voltage'],**{'voltage':[" IS NULL"],})

def mainRoads(osm_path):
    """
    Function to extract main road linestrings from OpenStreetMap    
    Arguments:
        *osm_path* : file path to the .osm.pbf file of the region 
        for which we want to do the analysis.        
    Returns:
        *GeoDataFrame* : a geopandas GeoDataFrame with all unique main road linestrings.   
    """ 
    return retrieve(osm_path,'lines',['highway','oneway','lanes','maxspeed'],**{'highway':["='primary' or ","='trunk' or ","='motorway' or ","='trunk_link' or ",
                    "='primary_link' or ", "='secondary' or ","='tertiary' or ","='tertiary_link'"]})


def remove_overlap_openstreetmap(gdf):
    """
    Function to remove overlap in polygons in from OpenStreetMap.
    
    Arguments:
        *gdf* : a geopandas GeoDataFrame with all unique railway linestrings.
        
    Returns:
        *GeoDataFrame* : a geopandas GeoDataFrame with (almost) non-overlapping polygons.
    
    """
    
    gdf['sq_area'] = gdf.area

    new_landuse = []
    for use in tqdm(gdf.itertuples(index=False),total=len(gdf),desc='Get unique shapes'):
        use_geom = use.geometry
        matches = gdf.loc[list(gdf.sindex.intersection(use.geometry.bounds))]
        for match in matches.itertuples(index=False):
            if use.sq_area > match.sq_area:
                use_geom = use_geom.difference(match.geometry)
        new_landuse.append([use.osm_id,use.landuse,use_geom])

    new_gdf  =  geopandas.GeoDataFrame(pandas.DataFrame(new_landuse,columns=['osm_id','landuse','geometry'])) 
    new_gdf.crs = {'init' : 'epsg:4326'}
    return new_gdf


def extract_value_other_gdf(x,gdf,col_name):
    """
    Function to extract value from column from other GeoDataFrame
    
    Arguments:
        *x* : row of main GeoDataFrame.
        
        *gdf* : geopandas GeoDataFrame from which we want to extract values.
        
        *col_name* : the column name from which we want to get the value.
        
    
    """
    try:
        return gdf.loc[list(gdf.sindex.intersection(x.geometry.bounds))][col_name].values[0]
    except:
        return None

def reproject(df_ds,current_crs="epsg:4326",approximate_crs = "epsg:3035"):
    """_summary_

    Args:
        df_ds (_type_): _description_
        current_crs (str, optional): _description_. Defaults to "epsg:4326".
        approximate_crs (str, optional): _description_. Defaults to "epsg:3035".

    Returns:
        _type_: _description_
    """    
    geometries = df_ds['geometry']
    coords = shapely.get_coordinates(geometries)
    transformer=pyproj.Transformer.from_crs(current_crs, approximate_crs,always_xy=True)
    new_coords = transformer.transform(coords[:, 0], coords[:, 1])
    
    return shapely.set_coordinates(geometries.copy(), np.array(new_coords).T) 
    
def get_damage_per_object(obj,df_ds,objects,curves,maxdam):
    """_summary_

    Args:
        obj (): The object for which we want to calculate the damage.
        df_ds (pandas DataFrame): The dataframe with the hazard points.
        objects (pandas DataFrame): The dataframe with the objects.
        curves (pandas DataFrame): The dataframe with the vulnerability curves.
        maxdam (dictionary): The dictionary with the maximum damage per object type.

    Returns:
        _type_: The damage per object.
    """  

    # find the exact hazard overlays:
    get_hazard_points = df_ds.iloc[obj[1]['hazard_point'].values].reset_index()
    get_hazard_points = get_hazard_points.loc[shapely.intersects(get_hazard_points.geometry.values,objects.iloc[obj[0]].geometry)]

    object_type = objects.iloc[obj[0]].obj_type
    object_geom = objects.iloc[obj[0]].geometry
    
    maxdam_object = maxdam[object_type]

    hazard_intensity = curves[object_type].index.values
    fragility_values = curves[object_type].values
                
    if len(get_hazard_points) == 0:
        return obj[0],0
    else:
        
        if shapely.get_type_id(object_geom) == 1:
            get_hazard_points['overlay_meters'] = shapely.length(shapely.intersection(get_hazard_points.geometry.values,object_geom))
            return obj[0],np.sum((np.interp(get_hazard_points.haz_val.values,hazard_intensity,fragility_values))*get_hazard_points.overlay_meters*maxdam_object)
        
        elif (shapely.get_type_id(object_geom) == 3) | (shapely.get_type_id(object_geom) == 6):
            get_hazard_points['overlay_m2'] = shapely.area(shapely.intersection(get_hazard_points.geometry.values,object_geom))
            return obj[0],get_hazard_points.apply(lambda x: np.interp(x.haz_val, hazard_intensity, fragility_values)*maxdam_object*x.overlay_m2,axis=1).sum()     
        
        else:
            print(shapely.get_type_id(object_geom))
            return obj[0],np.sum((np.interp(get_hazard_points.haz_val.values,hazard_intensity,fragility_values))*maxdam_object)


"""DamageScanner - a directe damage assessment toolkit

Raster specific functions
"""

from osgeo import gdal


def match(raster_in1,raster_in2):
    """
    In case of a mismatch between two rasters, return only the intersecting parts.
    
    Code adapted from http://sciience.tumblr.com/post/101722591382/finding-the-georeferenced-intersection-between-two
    
    Arguments:
        *raster_in1* : One of the two rasters to be clipped to the overlapping extent.
        
        *raster_in2* : One of the two rasters to be clipped to the overlapping extent.
        
    Returns:
        *array1* : Numpy Array of raster1
        
        *array2* : Numpy Array of raster2
        
        *intersection* : Bounding box of overlapping part
    
    """
    # Read rasterdata
    raster1 = gdal.Open(raster_in1)
    raster2 = gdal.Open(raster_in2)
    
    # load data
    band1 = raster1.GetRasterBand(1)
    band2 = raster2.GetRasterBand(1)
    gt1 = raster1.GetGeoTransform()
    gt2 = raster2.GetGeoTransform()
    
    # find each image's bounding box
    # r1 has left, top, right, bottom of dataset's bounds in geospatial coordinates.
    r1 = [gt1[0], gt1[3], gt1[0] + (gt1[1] * raster1.RasterXSize), gt1[3] + (gt1[5] * raster1.RasterYSize)]
    r2 = [gt2[0], gt2[3], gt2[0] + (gt2[1] * raster2.RasterXSize), gt2[3] + (gt2[5] * raster2.RasterYSize)]
    print('\t1 bounding box: %s' % str(r1))
    print('\t2 bounding box: %s' % str(r2))
    
    # find intersection between bounding boxes
    intersection = [max(r1[0], r2[0]), min(r1[1], r2[1]), min(r1[2], r2[2]), max(r1[3], r2[3])]
    if r1 != r2:
        print('\t** different bounding boxes **')
        # check for any overlap at all...
        if (intersection[2] < intersection[0]) or (intersection[1] < intersection[3]):
            intersection = None
            print('\t***no overlap***')
            return
        else:
            print('\tintersection:',intersection)
            left1 = int(round((intersection[0]-r1[0])/gt1[1])) # difference divided by pixel dimension
            top1 = int(round((intersection[1]-r1[1])/gt1[5]))
            col1 = int(round((intersection[2]-r1[0])/gt1[1])) - left1 # difference minus offset left
            row1 = int(round((intersection[3]-r1[1])/gt1[5])) - top1
            
            left2 = int(round((intersection[0]-r2[0])/gt2[1])) # difference divided by pixel dimension
            top2 = int(round((intersection[1]-r2[1])/gt2[5]))
            col2 = int(round((intersection[2]-r2[0])/gt2[1])) - left2 # difference minus new left offset
            row2 = int(round((intersection[3]-r2[1])/gt2[5])) - top2
            
            #print '\tcol1:',col1,'row1:',row1,'col2:',col2,'row2:',row2
            if col1 != col2 or row1 != row2:
                print("ERROR: Columns and rows still do not match! ***")
            # these arrays should now have the same spatial geometry though NaNs may differ
            array1 = band1.ReadAsArray(left1,top1,col1,row1)
            array2 = band2.ReadAsArray(left2,top2,col2,row2)

    else: # same dimensions from the get go
        col1 = raster1.RasterXSize # = col2
        row1 = raster1.RasterYSize # = row2
        array1 = band1.ReadAsArray()
        array2 = band2.ReadAsArray()
        
    return array1, array2, intersection


from setuptools import setup

def long_description():
    with open('README.md', encoding='utf-8') as f:
        return f.read()

setup(
    name='damagescanner',    
    version='0.6.1',
    long_description=long_description(),
    long_description_content_type='text/markdown',    
    install_requires=[
    "geopandas",
    "matplotlib",	
    "numpy",
    "pyproj",
    "rasterio",
    "xarray",
    "rioxarray",
    "packaging",
    "pandas",
    "shapely",
    ],
    project_urls={
        'GitHub': 'https://github.com/ElcoK/DamageScanner'
    },

)


"""DamageScanner - a directe damage assessment toolkit

Copyright (C) 2019 Elco Koks. All versions released under the MIT license.
"""

# Get all the needed modules
import os
import rasterio
import xarray as xr
import numpy as np
import shapely
import pandas as pd
import geopandas as gpd
from affine import Affine
import pyproj
from tqdm import tqdm
from os.path import join as p_join
import warnings


from damagescanner.vector import reproject,get_damage_per_object
from damagescanner.raster import match


def check_output_path(given_args):
    """Ensures given output path exists.

    Arguments:
        *given_args* : dict, of keyword arguments.

    Returns:
        *str* : output_path, which may be empty string ('')
    """
    output_path = given_args.get('output_path', '')
    if output_path != '' and not os.path.exists(output_path):
        os.mkdir(output_path)
    return output_path


def check_scenario_name(given_args):
    """Ensures given output path exists.

    Arguments:
        *given_args* : dict, of keyword arguments.

    Returns:
        *str* : scenario_name
    """
    scenario_name = given_args.get('scenario_name', False)
    if not scenario_name:
        raise ValueError("Required `scenario_name` not defined.")

    return scenario_name


def RasterScanner(landuse_file,
                  hazard_file,
                  curve_path,
                  maxdam_path,
                  lu_crs=28992,
                  haz_crs=4326,                   
                  hazard_col='FX',                  
                  dtype = np.int32,
                  save=False,
                  **kwargs):
    """
    Raster-based implementation of a direct damage assessment.
    
    Arguments:
        *landuse_file* : GeoTiff with land-use information per grid cell. Make sure 
        the land-use categories correspond with the curves and maximum damages 
        (see below). Furthermore, the resolution and extend of the land-use map 
        has to be exactly the same as the inundation map.
     
        *hazard_file* : GeoTiff or netCDF4 with hazard intensity per grid cell. Make sure 
        that the unit of the hazard map corresponds with the unit of the 
        first column of the curves file.
     
        *curve_path* : File with the stage-damage curves of the different 
        land-use classes. Values should be given as ratios, i.e. between 0 and 1.
        Can also be a pandas DataFrame or numpy Array.
     
        *maxdam_path* : File with the maximum damages per land-use class 
        (in euro/m2). Can also be a pandas DataFrame or numpy Array.
                        
        *dtype*: Set the dtype to the requires precision. This will affect the output damage raster as well 
     
    Optional Arguments:
        *save* : Set to True if you would like to save the output. Requires 
        several **kwargs**
        
    kwargs:
        *nan_value* : if nan_value is provided, will mask the inundation file. 
        This option can significantly fasten computations
        
        *cell_size* : If both the landuse and hazard map are numpy arrays, 
        manually set the cell size.

        *resolution* : If landuse is a numpy array, but the hazard map 
        is a netcdf, you need to specify the resolution of the landuse map.

        *output_path* : Specify where files should be saved.
        
        *scenario_name*: Give a unique name for the files that are going to be saved.
        
        *in_millions*: Set to True if all values should be set in millions.
        
        *crs*: Specify crs if you only read in two numpy array
        
        *transform*: Specify transform if you only read in numpy arrays in order to save the result raster
        
    Raises:
        *ValueError* : on missing kwarg options

    Returns:    
     *damagebin* : Table with the land-use class numbers (1st column) and the 
     damage for that land-use class (2nd column).
     
     *damagemap* : Map displaying the damage per grid cell of the area.
     
    """
    # load land-use map
    if isinstance(landuse_file, str):
        with rasterio.open(landuse_file) as src:
            landuse = src.read()[0, :, :]
            transform = src.transform
            resolution = src.res[0]
            cellsize = src.res[0] * src.res[1]
    else:
        landuse = landuse_file.copy()

    landuse_in = landuse.copy()
        
    # Load hazard map
    if isinstance(hazard_file, str):
        if (hazard_file.endswith('.tif') | hazard_file.endswith('.tiff')):      
            with rasterio.open(hazard_file) as src:
                hazard = src.read()[0, :, :]
                transform = src.transform
                
        elif hazard_file.endswith('.nc'):

            # Open the hazard netcdf file and store it in the hazard variable
            hazard = xr.open_dataset(hazard_file)

            # Set the crs of the hazard variable to haz_crs
            hazard.rio.write_crs(haz_crs, inplace=True)

            # Rename the latitude and longitude variables to 'y' and 'x' respectively
            hazard = hazard.rename({'Latitude': 'y','Longitude': 'x'})

            # Set the x and y dimensions in the hazard variable to 'x' and 'y' respectively
            hazard.rio.set_spatial_dims(x_dim="x",y_dim="y", inplace=True)

            # Open the landuse geotiff file and store it in the landuse variable
            landuse = xr.open_dataset(landuse_file, engine="rasterio")

            # Set the crs of the landuse variable to lu_crs
            landuse.rio.write_crs(lu_crs,inplace=True)

            # Reproject the landuse variable from EPSG:4326 to EPSG:3857
            landuse = landuse.rio.reproject("EPSG:3857",resolution=resolution)

            # Get the minimum longitude and latitude values in the landuse variable
            min_lon = landuse.x.min().to_dict()['data']
            min_lat = landuse.y.min().to_dict()['data']

            # Get the maximum longitude and latitude values in the landuse variable
            max_lon = landuse.x.max().to_dict()['data']
            max_lat = landuse.y.max().to_dict()['data']

            # Create a bounding box using the minimum and maximum latitude and longitude values
            area = gpd.GeoDataFrame([shapely.box(min_lon,min_lat,max_lon, max_lat)],columns=['geometry'])

            # Set the crs of the bounding box to EPSG:3857
            area.crs = 'epsg:3857'

            # Convert the crs of the bounding box to EPSG:4326
            area = area.to_crs('epsg:4326')

            # Clip the hazard variable to the extent of the bounding box
            hazard = hazard.rio.clip(area.geometry.values, area.crs)

            # Reproject the hazard variable to EPSG:3857 with the desired resolution
            hazard = hazard.rio.reproject("EPSG:3857",resolution=resolution)

            # Clip the hazard variable again to the extent of the bounding box
            hazard = hazard.rio.clip(area.geometry.values, area.crs)

            # If the hazard variable has fewer columns and rows than the landuse variable, reproject the landuse variable to match the hazard variable
            if (len(hazard.x)<len(landuse.x)) & (len(hazard.y)<len(landuse.y)):
                landuse= landuse.rio.reproject_match(hazard)

            # If the hazard variable has more columns and rows than the landuse variable, reproject the hazard variable to match the landuse variable

            elif (len(hazard.x)>len(landuse.x)) & (len(hazard.y)>len(landuse.y)):
                hazard = hazard.rio.reproject_match(landuse)

            # Convert the hazard and landuse variable to a numpy array
            landuse = landuse['band_data'].to_numpy()[0,:,:]
            hazard = hazard[hazard_col].to_numpy()[0,:,:]
        
        else:
            hazard = hazard_file.copy()

    # check if land-use and hazard map have the same shape.
    if landuse.shape != hazard.shape:
        warnings.warn(
            "WARNING: landuse and hazard maps are not the same shape. Let's fix this first!"
        )
        
        landuse, hazard, intersection = match(landuse_file, hazard_file)

        # create the right affine for saving the output
        transform = Affine(transform[0], transform[1], intersection[0],
                           transform[3], transform[4], intersection[1])

    # set cellsize:
    if isinstance(landuse_file, str) | isinstance(hazard_file, str):
        cellsize = src.res[0] * src.res[1]
    else:
        try:
            cellsize = kwargs['cellsize']
        except KeyError:
            raise ValueError("Required `cellsize` not given.")

    # Load curves
    if isinstance(curve_path, pd.DataFrame):
        curves = curve_path.values
    elif isinstance(curve_path, np.ndarray):
        curves = curve_path
    elif curve_path.endswith('.csv'):
        curves = pd.read_csv(curve_path).values
    
    if ((curves>1).all()) or ((curves<0).all()):
        raise ValueError("Stage-damage curve values must be between 0 and 1")

    # Load maximum damages
    if isinstance(maxdam_path, pd.DataFrame):
        maxdam = maxdam_path.values
    elif isinstance(maxdam_path, np.ndarray):
        maxdam = maxdam_path
    elif maxdam_path.endswith('.csv'):
        maxdam = pd.read_csv(maxdam_path).values
    
    if maxdam.shape[0] != (curves.shape[1]-1):
        raise ValueError("Dimensions between maximum damages and the number of depth-damage curve do not agree")
  
    # Speed up calculation by only considering feasible points
    if kwargs.get('nan_value'):
        nan_value = kwargs.get('nan_value')
        hazard[hazard == nan_value] = 0

    haz = hazard * (hazard >= 0) + 0   
    haz[haz >= curves[:, 0].max()] = curves[:, 0].max()   
    area = haz > 0
    haz_intensity = haz[haz > 0]
    landuse = landuse[haz > 0]

    # Calculate damage per land-use class for structures
    numberofclasses = len(maxdam)
    alldamage = np.zeros(landuse.shape[0])
    damagebin = np.zeros((
        numberofclasses,
        2,
    ))
    for i in range(0, numberofclasses):
        n = maxdam[i, 0]
        damagebin[i, 0] = n
        wd = haz_intensity[landuse == n]
        alpha = np.interp(wd, ((curves[:, 0])), curves[:, i + 1])
        damage = alpha * (maxdam[i, 1] * cellsize)
        damagebin[i, 1] = sum(damage)
        alldamage[landuse == n] = damage
    
    # create the damagemap
    damagemap = np.zeros((area.shape[0], area.shape[1]), dtype=dtype)
    damagemap[area] = alldamage

    # create pandas dataframe with output
    damage_df = pd.DataFrame(damagebin.astype(dtype),
                           columns=['landuse',
                                    'damages']).groupby('landuse').sum()

    if save:
        crs = kwargs.get('crs', src.crs)
        transform = kwargs.get('transform', transform)

        # requires adding output_path and scenario_name to function call
        # If output path is not defined, will place file in current directory
        output_path = check_output_path(kwargs)
        scenario_name = check_scenario_name(kwargs)
        path_prefix = p_join(output_path, scenario_name)

        damage_fn = '{}_damages.csv'.format(path_prefix)
        damage_df.to_csv(damage_fn)

        dmap_fn = '{}_damagemap.tif'.format(path_prefix)
        rst_opts = {
            'driver': 'GTiff',
            'height': damagemap.shape[0],
            'width': damagemap.shape[1],
            'count': 1,
            'dtype': dtype,
            'crs': crs,
            'transform': transform,
            'compress': "LZW"
        }
        with rasterio.open(dmap_fn, 'w', **rst_opts) as dst:
            dst.write(damagemap, 1)

    if 'in_millions' in kwargs:
        damage_df = damage_df / 1e6

    # return output
    return damage_df, damagemap, landuse_in, hazard


def VectorScanner(exposure_file,
                  hazard_file,
                  curve_path,
                  maxdam_path,
                  cell_size = 5,
                  exp_crs=4326,
                  haz_crs=4326,                   
                  object_col='landuse',
                  hazard_col='inun_val',
                  centimeters=False,
                  save=False,
                  **kwargs):
    """
    Vector based implementation of a direct damage assessment
    
    Arguments:
        *exposure_file* : Shapefile, Pandas DataFrame or Geopandas GeoDataFrame 
        with land-use information of the area.
     
        *hazard_file* : GeoTiff with inundation depth per grid cell. Make sure 
        that the unit of the inundation map corresponds with the unit of the 
        first column of the curves file. 
     
        *curve_path* : File with the stage-damage curves of the different 
        land-use classes. Can also be a pandas DataFrame (but not a numpy Array).
     
        *maxdam_path* : File with the maximum damages per land-use class 
        (in euro/m2). Can also be a pandas DataFrame (but not a numpy Array).

    Optional Arguments:
        *cell_size* : Specify the cell size of the hazard map. 

        *exp_crs* : Specify the coordinate reference system of the exposure. Default is 
        set to **4326**. A preferred CRS system is in meters. Please note that the function
        only accepts the CRS system in the form of an integer EPSG code.
        
        *haz_crs* : Specify the cooordinate reference system of the hazard. Default is 
        set to **4326**.  A preferred CRS system is in meters.Please note that the function
        only accepts the CRS system in the form of an integer EPSG code.
        
        *centimeters* : Set to True if the inundation map and curves are in 
        centimeters
        
        *object_col* : Specify the column name of the unique object id's. 
        Default is set to **landuse**.
        
        *hazard_col* : Specify the column name of the hazard intensity 
        Default is set to **inun_val**.       
        
        *save* : Set to True if you would like to save the output. Requires 
        several **kwargs**
        
    kwargs:
        *output_path* : Specify where files should be saved.
        
        *scenario_name*: Give a unique name for the files that are going to be saved.
        
        *print_tqdm*: Set to **False** when progress output is undesired.
    
    Raises:
        *ValueError* : on missing kwargs
    
    Returns:    
     *damagebin* : Table with the land-use class names (1st column) and the 
     damage for that land-use class (2nd column).
     
    """
    # load exposure data
    if isinstance(exposure_file, str):
        exposure = gpd.read_file(exposure_file)
    
    elif (isinstance(exposure_file, gpd.GeoDataFrame) | isinstance(exposure_file, pd.DataFrame)):
        exposure = gpd.GeoDataFrame(exposure_file.copy())

    else:
        print(
            'ERROR: exposure data should either be a shapefile, GeoPackage, a GeoDataFrame or a pandas Dataframe with a geometry column'
        )

    # load hazard file
    if isinstance(hazard_file, str):       
        if (hazard_file.endswith('.tif') | hazard_file.endswith('.tiff')): 
            
            #load dataset
            hazard_map = xr.open_dataset(hazard_file, engine="rasterio")
            
            #specify hazard_col
            hazard_col = 'band_data'
            
            #convert to dataframe
            hazard = hazard_map['band_data'].to_dataframe().reset_index()
            
            # drop all non values and zeros to reduce size
            hazard = hazard.loc[~(hazard['band_data'].isna() 
                                                      | hazard['band_data']<=0)].reset_index(drop=True)
                    
            # create geometry values and drop lat lon columns
            hazard['geometry'] = shapely.points(np.array(list(zip(hazard['x'],hazard['y']))))
            hazard = hazard.drop(['x','y','band','spatial_ref'],axis=1)

            #and turn them into squares again:
            hazard.geometry= shapely.buffer(hazard.geometry,
                                               distance=cell_size/2,cap_style='square').values
            
        elif hazard_file.endswith('.nc'):
           #load dataset
            hazard_map = xr.open_dataset(hazard_file)
            
            #convert to dataframe
            hazard = hazard_map[hazard_col].to_dataframe().reset_index()
            
            # drop all non values and below zeros to reduce size. Might cause issues for cold waves
            hazard = hazard.loc[~(hazard[hazard_col].isna() | hazard[hazard_col]<=0)].reset_index(drop=True)
            
           # create geometry values and drop lat lon columns
            hazard['geometry'] = shapely.points(np.array(list(zip(hazard['x'],hazard['y']))))
            hazard = hazard.drop(['x','y','band','spatial_ref'],axis=1)

            #and turn them into squares again:
            hazard.geometry= shapely.buffer(hazard.geometry,
                                               distance=cell_size/2,cap_style='square').values
     


        elif (hazard_file.endswith('.shp') | hazard_file.endswith('.gpkg')):
            hazard = gpd.read_file(hazard_file)
        else:
            print('ERROR: hazard data should either be a GeoTIFF, a netCDF4, a shapefile or a GeoPackage.'
                 )          

    elif (isinstance(hazard_file, gpd.GeoDataFrame) | isinstance(hazard_file, pd.DataFrame)):
        hazard = gpd.GeoDataFrame(hazard_file.copy())
    else:
        raise ValueError(
            'ERROR: hazard data should be a GeoTiff, a netCDF4, a shapefile, a GeoDataFrame \
              or any other georeferenced format that can be read by xarray or geopandas'
        )

    print("PROGRESS: Exposure and hazard data loaded")

    #check if exposure and hazard data are in the same CRS and in a CRS in meters
    CRS_exposure = pyproj.CRS.from_epsg(exp_crs)
    CRS_hazard = pyproj.CRS.from_epsg(haz_crs)

    exp_crs_unit_name = CRS_exposure.axis_info[0].unit_name
    haz_crs_unit_name = CRS_hazard.axis_info[0].unit_name

    if exp_crs_unit_name == 'degree' and haz_crs_unit_name == 'metre':
        exposure.geometry = reproject(exposure,current_crs=f"epsg:{exp_crs}",approximate_crs = f"epsg:{haz_crs}")
    elif exp_crs_unit_name == 'degree' and haz_crs_unit_name == 'degree':
         exposure.geometry = reproject(exposure,current_crs=f"epsg:{exp_crs}",approximate_crs = "epsg:3857")  

    if haz_crs_unit_name == 'degree' and exp_crs_unit_name == 'metre':
        hazard.geometry = reproject(hazard,current_crs=haz_crs,approximate_crs = f"epsg:{exp_crs}")
    elif haz_crs_unit_name == 'degree' and exp_crs_unit_name == 'degree':
        hazard.geometry = reproject(hazard,current_crs=haz_crs,approximate_crs = "epsg:3857")
         
    # rename inundation colum, set values to centimeters if required and to integers:
    hazard = hazard.rename(columns={hazard_col:'haz_val'})
    if not centimeters:
        hazard['haz_val'] = hazard.haz_val*100
        hazard['haz_val'] = hazard.haz_val.astype(int)

    # rename object col to make sure everything is consistent:
    exposure = exposure.rename({object_col:'obj_type'},axis=1)
            
    # Load curves
    if isinstance(curve_path, pd.DataFrame):
        curves = curve_path.copy()
    elif isinstance(curve_path, np.ndarray):
        raise ValueError(
            'ERROR: for the vector-based approach we use a pandas DataFrame, not a Numpy Array'
        )
    elif curve_path.endswith('.csv'):
        curves = pd.read_csv(curve_path, index_col=[0])

    # Load maximum damages
    if isinstance(maxdam_path, str) and maxdam_path.endswith('.csv'):
        maxdam_path = pd.read_csv(maxdam_path)
    elif isinstance(maxdam_path, pd.DataFrame):
        maxdam = dict(zip(maxdam_path[object_col], maxdam_path['damage']))
    elif isinstance(maxdam_path, np.ndarray):
        maxdam = dict(zip(maxdam_path[:, 0], maxdam_path[:, 1]))
    elif isinstance(maxdam_path, dict):
        maxdam = maxdam_path

    # Check if we need to turn off tqdm:
    tqdm_print = kwargs.get('print_tqdm', True)

    #overlay hazard and exposure data
    hazard_tree = shapely.STRtree(hazard.geometry.values)
    
    if (shapely.get_type_id(exposure.iloc[0].geometry) == 3) | (shapely.get_type_id(exposure.iloc[0].geometry) == 6):
        overlay = hazard_tree.query(exposure.geometry,predicate='intersects')    
    else:
        overlay = hazard_tree.query(shapely.buffer(exposure.geometry.values,distance=cell_size*2),predicate='intersects')
    
    overlay_exp_haz = pd.DataFrame(overlay.T,columns=['obj_type','hazard_point'])
    
    #perform calculation
    collect_output = []
    
    for obj in tqdm(overlay_exp_haz.groupby('obj_type'),total=len(overlay_exp_haz.obj_type.unique()),
                                  desc='damage calculation'):
        collect_output.append(get_damage_per_object(obj,hazard,exposure,curves,maxdam))
        
    damaged_objects = exposure.merge(pd.DataFrame(collect_output,columns=['index','damage']),
                                                          left_index=True,right_on='index')[['obj_type','geometry','damage']]


    if save == True:
        # requires adding output_path and scenario_name to function call
        # If output path is not defined, will place file in current directory
        output_path = check_output_path(kwargs)
        scenario_name = check_scenario_name(kwargs)
        path_prefix = p_join(output_path, scenario_name)

        damage_fn = f'{path_prefix}_damages.csv'
        damaged_objects.to_csv(damage_fn)
        return damaged_objects
        
    else:
        return damaged_objects 
